#ifndef SMALLGMP_H
#define SMALLGMP_H
/* smallgmp.h -- Definitions for multiple precision functions,
used in SINGULAR and compatible with GNU MP Library. */

/* Allow faster testing for negative, zero, and positive.  */
#define mpz_sgn(Z) ((Z)->_mp_size < 0 ? -1 : (Z)->_mp_size > 0)

typedef unsigned int            mp_limb_t;
typedef mp_limb_t               mp_limb;

typedef mp_limb_t *             mp_ptr;
typedef const mp_limb_t *       mp_srcptr;
typedef struct
{
  int _mp_alloc;	     /* Number of *limbs* allocated and pointed
                                to by the D field.  */
  int _mp_size; 	     /* abs(SIZE) is the number of limbs
                                the last field points to.  If SIZE
                                is negative this is a negative
                                number.  */
  mp_limb_t *_mp_d;          /* Pointer to the limbs.  */
} MP_INT;
typedef MP_INT * mpz_ptr;
extern const int mp_bits_per_limb;

#include <stdio.h>
#include <stdlib.h>

#ifndef SMALLGMP_DEBUG
#define mmpAlloc(size) ((*_mp_allocate_func)(size))
#define mmpFree(adr,size) ((*_mp_free_func)(adr,size))

extern void * (*_mp_allocate_func) (size_t);
extern void * (*_mp_reallocate_func) (void *, size_t, size_t);
extern void (*_mp_free_func) (void *, size_t);
void mp_set_memory_functions (void *(*) (size_t),
     void *(*) (void *, size_t, size_t),
     void (*) (void *, size_t));
#else
#define mmpAlloc(size) ((*_mp_allocate_func)(size,__FILE__,__LINE__))
#define mmpFree(adr,size) ((*_mp_free_func)(adr,size,__FILE__,__LINE__))

extern void * (*_mp_allocate_func) (size_t,char*,int);
extern void * (*_mp_reallocate_func) (void *, size_t, size_t,char*,int);
extern void (*_mp_free_func) (void *, size_t,char*,int);
void mp_set_memory_functions (void *(*) (size_t,char*,int),
     void *(*) (void *, size_t, size_t,char*,int),
     void (*) (void *, size_t,char*,int));
#endif

void mpz_init (MP_INT *);
void _mpz_realloc (MP_INT *, int);
void mpz_init_set (MP_INT *, const MP_INT *);
void mpz_set (MP_INT *, const MP_INT *);
void mpz_init_set_ui (MP_INT *, unsigned long int);
void mpz_init_set_si (MP_INT *, signed long int);
void mpz_set_si (MP_INT *, signed long int);
signed long int mpz_get_si (const MP_INT *);
void mpz_clear (MP_INT *);
void mpz_set_str (MP_INT *, const char *, int);
size_t mpz_sizeinbase (const MP_INT *, int);
char * mpz_get_str (char *, int, const MP_INT *);

int mpz_cmp (const MP_INT *, const MP_INT *);
int mpz_cmp_ui (const MP_INT *, unsigned long int);
int mpz_cmp_si (const MP_INT *, signed long int);

void mpz_neg (MP_INT *, const MP_INT *);
void mpz_add (MP_INT *, const MP_INT *, const MP_INT *);
void mpz_add_ui (MP_INT *, const MP_INT *, unsigned long int);
void mpz_sub (MP_INT *, const MP_INT *, const MP_INT *);
void mpz_sub_ui (MP_INT *, const MP_INT *, unsigned long int);
void mpz_mul (MP_INT *, const MP_INT *, const MP_INT *);
void mpz_mul_ui (MP_INT *, const MP_INT *, unsigned long int);

void mpz_tdiv_qr (MP_INT *, MP_INT *, const MP_INT *, const MP_INT *);
void mpz_fdiv_qr (MP_INT *, MP_INT *, const MP_INT *, const MP_INT *);
void mpz_cdiv_qr (MP_INT *, MP_INT *, const MP_INT *, const MP_INT *);
void mpz_tdiv_q (MP_INT *, const MP_INT *, const MP_INT *);
void mpz_fdiv_q (MP_INT *, const MP_INT *, const MP_INT *);
void mpz_cdiv_q (MP_INT *, const MP_INT *, const MP_INT *);
void mpz_tdiv_r (MP_INT *, const MP_INT *, const MP_INT *);
void mpz_mod (MP_INT *, const MP_INT *, const MP_INT *);
unsigned long int mpz_fdiv_ui (const MP_INT *, unsigned long int);
unsigned long int mpz_cdiv_ui (const MP_INT *, unsigned long int);
void mpz_fdiv_r_ui (MP_INT *, const MP_INT *, unsigned long int);
unsigned long int mpz_fdiv_q_ui (MP_INT *, const MP_INT *, unsigned long int);
unsigned long int mpz_fdiv_qr_ui (MP_INT *, MP_INT *, const MP_INT *, unsigned long int);
void mpz_gcd (MP_INT *, const MP_INT *, const MP_INT *);
unsigned long int 
mpz_gcd_ui (MP_INT *w, const MP_INT *u, unsigned long int v);
void mpz_pow_ui (MP_INT *, const MP_INT *, unsigned long int);

/* Factory */
void mpz_init_set_str (MP_INT *, const char *, int);
void mpz_set_ui (MP_INT *, unsigned long int);
void mpz_sqrt (MP_INT *, const MP_INT *);
void mpz_gcdext (MP_INT *, MP_INT *, MP_INT *, const MP_INT *, const MP_INT *);

unsigned long int mpz_get_ui (const MP_INT *);
void mpz_abs (MP_INT *, const MP_INT *);


/* factory */
#define mpz_divexact    mpz_tdiv_q
/*   Compatibility with GMP 1. */
#define mpz_mdiv	mpz_fdiv_q
#define mpz_mdivmod	mpz_fdiv_qr
#define mpz_mmod	mpz_fdiv_r
#define mpz_mdiv_ui	mpz_fdiv_q_ui
#define mpz_mdivmod_ui(q,r,n,d) ((r == 0) ? mpz_fdiv_q_ui (q,n,d) : mpz_fdiv_qr_ui (q,r,n,d))
#define mpz_mmod_ui(r,n,d) mpz_fdiv_ui (n,d)

/*  Useful synonyms, but not quite compatible with GMP 1. */
#define mpz_div		mpz_fdiv_q
#define mpz_divmod	mpz_fdiv_qr
#define mpz_div_ui	mpz_fdiv_q_ui
#define mpz_divmod_ui	mpz_fdiv_qr_ui
#define mpz_mod_ui	mpz_fdiv_r_ui
#define mpz_div_2exp	mpz_fdiv_q_2exp
#define mpz_mod_2exp	mpz_fdiv_r_2exp

#endif

