#include <stuff_64.h>

#ifdef __MWERKS__
#ifdef macintosh
typedef union
{
  unsigned long long x;
  unsigned y[2];
} ll_convert;
void mul_64(INT64 rr, cu_int aa, cu_int bb)
{
  ll_convert a, r;
  a.y[HIGH] = r.y[HIGH] = 0;
  a.y[LOW] = aa;
  r.y[LOW] = bb;
  r.x *= a.x;
  rr->lo = r.y[LOW];
  rr->hi = r.y[HIGH];
}
#else
extern void mul_64(INT64 rr, cu_int aa, cu_int bb)
{
  asm
  {
    mov eax, aa
    mul bb
    mov ebx, eax
    mov eax, rr
    mov 0[eax], ebx
    mov 4[eax], edx
  }
}
#endif
#else
void mul_64(INT64 rr, cu_int aa, cu_int bb)
{
  unsigned long long a=aa, b=bb, r=a*b;
  rr->lo=r;
  rr->hi=r>>32;
}
#endif

u_int rem_64( const INT64 a, cu_int b )
{
  u_int hi = a->hi, lo = a->lo;
  u_int res, inc;
  
  if (hi >= b)
    hi %= b;
  if (!hi)
    return lo%b;
  if (b<0x10000)
    return rem_ll_s(hi, lo, b);
  else
  {
    inc = lead_zero(b);
    if (inc)
    {
      hi <<= inc;
      hi += lo>>(32-inc);
      lo <<= inc;
      res = rem_ll_u(hi, lo, b<<inc);
      if (res)
        res >>= inc;
    }
    else
      res = rem_ll_u(hi, lo, b);
    return res;
  }
}

u_int rem_ll_s( cu_int hi, cu_int lo, cu_int b)
{
  ii_convert x, y;
  
  y.u = lo;
  x.s[HIGH] = hi;
  x.s[LOW] = y.s[HIGH];
  y.s[HIGH] = x.u%b;
  return y.u%b;
}

u_int divmod_ll_s( u_int *rq, cu_int hi, cu_int lo, cu_int b)
{
  u_int q;
  ii_convert x, y, c;
  
  y.u = lo;
  x.s[HIGH] = hi;
  x.s[LOW] = y.s[HIGH];
  c.s[HIGH] = q = x.u/b;
  y.s[HIGH] = x.u-q*b;
  c.s[LOW] = q = y.u/b;
  *rq = c.u;
  return y.u-q*b;
}

u_int rem_ll_u( cu_int hi, cu_int lo, cu_int b)
{
  u_int div, quot;
  ii_convert x, y, z, w1, w2;
  
  x.u = hi;
  y.u = lo;
  z.u = b;
  div = z.s[HIGH]+1;
  while (x.s[HIGH])
  {
    quot = x.u/div;
    w1.u = quot*z.s[LOW];
    w2.u = quot*z.s[HIGH];
    if (w1.s[LOW]<=y.s[HIGH])
    {
      y.s[HIGH] -= w1.s[LOW];
      x.u -= (w2.u+w1.s[HIGH]); 
    }
    else
    {
      y.s[HIGH] -= w1.s[LOW];
      x.u -= (w2.u+w1.s[HIGH]+1); 
    }
  }
  w2.s[HIGH] = x.s[LOW];
  w2.s[LOW] = y.s[HIGH];
  if (w2.u>=z.u)
    w2.u -= z.u;
  while (w2.s[HIGH])
  {
    quot = w2.u/div;
    w1.u = quot*z.s[LOW];
    x.u = quot*z.s[HIGH];
    if (w1.s[LOW]<=y.s[LOW])
    {
      y.s[LOW] -= w1.s[LOW];
      w2.u -= (x.u+w1.s[HIGH]);
    }
    else
    {
      y.s[LOW] -= w1.s[LOW];
      w2.u -= (x.u+w1.s[HIGH]+1);
    }
  }
  y.s[HIGH] = w2.s[LOW];
  if (y.u>=z.u)
    y.u -= z.u;
  return y.u;
}

u_int divmod_ll_u( u_int *rq, cu_int hi, cu_int lo, cu_int b)
{
  u_int div, quot;
  ii_convert x, y, z, w1, w2, c;
  
  c.u = 0;
  x.u = hi;
  y.u = lo;
  z.u = b;
  div = z.s[HIGH]+1;
  while (x.s[HIGH])
  {
    quot = x.u/div;
    c.s[HIGH] += quot;
    w1.u = quot*z.s[LOW];
    w2.u = quot*z.s[HIGH];
    if (w1.s[LOW]<=y.s[HIGH])
    {
      y.s[HIGH] -= w1.s[LOW];
      x.u -= (w2.u+w1.s[HIGH]); 
    }
    else
    {
      y.s[HIGH] -= w1.s[LOW];
      x.u -= (w2.u+w1.s[HIGH]+1); 
    }
  }
  w2.s[HIGH] = x.s[LOW];
  w2.s[LOW] = y.s[HIGH];
  if (w2.u>=z.u)
  {
    w2.u -= z.u;
    c.s[HIGH]++;
  }
  while (w2.s[HIGH])
  {
    quot = w2.u/div;
    c.u += quot;
    w1.u = quot*z.s[LOW];
    x.u = quot*z.s[HIGH];
    if (w1.s[LOW]<=y.s[LOW])
    {
      y.s[LOW] -= w1.s[LOW];
      w2.u -= (x.u+w1.s[HIGH]);
    }
    else
    {
      y.s[LOW] -= w1.s[LOW];
      w2.u -= (x.u+w1.s[HIGH]+1);
    }
  }
  y.s[HIGH] = w2.s[LOW];
  if (y.u>=z.u)
  {
    y.u -= z.u;
    c.u++;
  }
  *rq = c.u;
  return y.u;
}

u_int div_ll_u(cu_int hi, cu_int lo, cu_int b)
{
  u_int div, quot;
  ii_convert x, y, z, w1, w2, c;
  
  if (!b)
    return hi;
  c.u = 0;
  x.u = hi;
  y.u = lo;
  z.u = b;
  div = z.s[HIGH]+1;
  while (x.s[HIGH])
  {
    quot = x.u/div;
    c.s[HIGH] += quot;
    w1.u = quot*z.s[LOW];
    w2.u = quot*z.s[HIGH];
    if (w1.s[LOW]<=y.s[HIGH])
    {
      y.s[HIGH] -= w1.s[LOW];
      x.u -= (w2.u+w1.s[HIGH]); 
    }
    else
    {
      y.s[HIGH] -= w1.s[LOW];
      x.u -= (w2.u+w1.s[HIGH]+1); 
    }
  }
  w2.s[HIGH] = x.s[LOW];
  w2.s[LOW] = y.s[HIGH];
  if (w2.u>=z.u)
  {
    w2.u -= z.u;
    c.s[HIGH]++;
  }
  while (w2.s[HIGH])
  {
    quot = w2.u/div;
    c.u += quot;
    w1.u = quot*z.s[LOW];
    x.u = quot*z.s[HIGH];
    if (w1.s[LOW]<=y.s[LOW])
    {
      y.s[LOW] -= w1.s[LOW];
      w2.u -= (x.u+w1.s[HIGH]);
    }
    else
    {
      y.s[LOW] -= w1.s[LOW];
      w2.u -= (x.u+w1.s[HIGH]+1);
    }
  }
  y.s[HIGH] = w2.s[LOW];
  if (y.u>=z.u)
    c.u++;
  return c.u;
}

u_int lead_zero(u_int b)
{
  u_int inc;
  
  if (b&0xff000000)
    inc = 0;
  else
  {
    inc = 8;
    b <<= 8;
  }
  if (!(b&0xf0000000))
  {
    inc += 4;
    b <<= 4;
  }
  while (!(b&0x80000000)) 
  {
    inc++;
    b <<= 1;
  }
  return inc;
}

u_int leading_zero(u_int b)
{
  u_int inc;
  
  if (b&0xffff0000)
    inc = 0;
  else
  {
    inc = 16;
    b <<= 16;
  }
  if (!(b&0xff000000))
  {
    inc += 8;
    b <<= 8;
  }
  if (!(b&0xf0000000))
  {
    inc += 4;
    b <<= 4;
  }
  while (!(b&0x80000000)) 
  {
    inc++;
    b <<= 1;
  }
  return inc;
}

